/*
 * =====================================================================================
 *
 *       Filename:  linklist.h
 *
 *    Description:  Link List Exercise with MYSQL Hooks
 *
 *        Version:  1.0
 *        Created:  03/21/2011 12:42:28 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ruben Safir, 
 *        Company:  NYLXS Workshops
 *
 * =====================================================================================
 */
#ifndef LINKLIST_H
#define LINKLIST_H 

#include<iostream>
#include<cstdlib>
namespace chainlist {


   /* 
    * =====================================================================================
    *        Class:  Node
    *  Description:  This is a node in the "List" Link List
    * =====================================================================================
    */

   template < class unk >
      class Node
      {
	 public:

	    // ====================  LIFECYCLE     =======================================
	    /* constructor      */
	    Node<unk>( unk value, Node<unk> *item_to_link_to = 0);
	    Node<unk>();
	    Node ( const Node &other ); /* copy constructor  - Need to write this! */ 
	    ~Node<unk> ();                          /* destructor       */


	    /* ====================  ACCESSORS     ======================================= */
	    inline unk* value ()const{
	       return value_; 
	    }
	    inline Node<unk> * next()const {
	       return next_;
	    }

	    inline void value(unk *);
	    inline void value(unk);
	    inline void next(Node<unk> *);

	    /* ====================  MUTATORS      ======================================= */

	    /* ====================  OPERATORS     ======================================= */

	    Node& operator=( const Node &other ); // assignment operator
	    unk& operator*(){ //gain access to real value
	       return *(value());/// test this out as an lvalue
	    }
	    unk& operator*(unk);

	 protected:
	    /* ====================  DATA MEMBERS  ======================================= */



	 private:
	    /* ====================  DATA MEMBERS  ======================================= */

	    unk * value_ ;
	    Node * next_;
      }; /* -----  end of template class Node  ----- */

   template<class unk>
      Node<unk>::Node(unk val, Node<unk> *item_to_link_to){
	 value(val);
	 if(!item_to_link_to){
	    next_ = 0;
	 }

	 else{
	    next(item_to_link_to->next());
	    item_to_link_to->next(this);
	 }
      }

   template<class unk>  //try to eliminate
      Node<unk>::Node():value_(0),next_(0){}


   template<class unk>
      Node<unk>::~Node(){
	 if(value_)
	    delete value_;
//	 std::cout << "Delete Node" << __LINE__ <<  std::endl;

      }

   template <typename unk>
      void Node<unk>::next(Node * next_node){
	 next_ = next_node;
      }	

   template<class unk>
      void Node<unk>::value(unk val){
	 value_ = new unk(val);
      }

   template<class unk>
      void Node<unk>::value(unk * val){
	 value_ = val;
      }

   /* ===================================================================================== */

   template < class T >
      class List
      {
	 public:

	    /* ====================  LIFECYCLE     ======================================= */
	    List<T>(): front_(0),end_(0), cursor_(0), size_(0){}                             /* constructor */
	    void remove_all();
	    ~List<T>();

	    /* ====================  ACCESSORS     ======================================= */
	    unsigned int size()const { return size_; }
	    inline Node<T>* const &front()const{ return front_;}
	    inline Node<T>* &endd(){return end_;}

	    inline Node<T> *& cursor(){return cursor_; }

	    inline void cursor(Node<T> * a){ cursor_ = a; }

	    inline void endd(Node<T> *a){ end_ = a; } 
	    inline void front(Node<T> *a){ front_ = a; } 
	    inline void display();



	    /* ====================  MUTATORS      ======================================= */

	    inline void sizeup(){  ++size_; }
	    inline void sizedown(){  --size_; }
	    void insert_front( T value); //add a node to the front 
	    void insert(T value, Node<T> * &place_after_this); //add a new node after the Node given and return the new node
	    void insert(T value); //add a new node at the end of the list
	    void insert(); //add a new node at the end of the list
	    void insert_end(T value); // Add a new node to the end

	    void find_value(T value); //find a value from it's first occurance in the list
	    Node<T>*& find_next_value(const T& value, Node<T>*& last_node_searched); //find a value in the list after the 
	    //current node which is not searched
	    void remove_node_by_value(T value);
	    void remove_node_by_node(Node<T> cur);
	    void remove_front();
	    void remove_end();
	    void sort(List<T>&);//shell sort



	    /* ====================  OPERATORS     ======================================= */

	    T& operator[](unsigned int const);





	 protected:
	    /* ====================  DATA MEMBERS  ======================================= */

	 private:
	    /* ====================  DATA MEMBERS  ======================================= */
	    Node<T> * front_;
	    Node<T> * end_;
	    Node<T> * cursor_;
	    unsigned int size_;

      }; /* ----------  end of template class Lists  ---------- */

   template <class T>
      void List<T>::insert_front ( T value )
      {
	 Node<T> *tmp = new Node<T>(value);
	 Node<T> * tmp2 = front(); 
	 if(tmp2 == 0){
	    //	tmp2 = tmp;
	    endd(tmp);//endd not null anymore
	    cursor(tmp);
	    tmp->next(0);
	    front(tmp);//front not null anymore
	 }
	 else{
	    tmp->next(tmp2);
	    front(tmp);
	    cursor(tmp); //curosr moved to the front
	 }

	 sizeup();

	 return ;
      }
   /* -----  end of template function insert_front  ----- */


   template < class T >
      void List<T>::insert ( T val, Node<T> *&node_in_front )
      {
	 //If there is no node sent, attach it to the end
	 Node<T> * tmp = new Node<T>(val);
	 if(!front()){
	    front(tmp);
	    endd(tmp);
	    cursor(tmp);
	    sizeup();
	    return;
	 }

	 if(!node_in_front){
	    endd()->next(tmp);
	    endd(tmp);
	    cursor(tmp); //Set Cursor to the end
	 }else{
	    tmp->next(node_in_front->next());
	    node_in_front->next(tmp);
	    cursor(tmp);
	    if(node_in_front == endd())
	       endd(tmp);
	 }
	 sizeup();

	 return ;
      }		/* -----  end of method List<T>::insert  ----- */

   template < class T >
      void List<T>::insert ( )
      {
	 if(!front()){
	    Node<T> * tmp = new Node<T>();
	    front(tmp);
	    endd(tmp);
	    cursor(tmp);
	    sizeup();
	    return;
	 }else{
	    Node<T> * tmp = new Node<T>();
	    endd()->next(tmp); //not redundant
	    endd(tmp);
	    cursor(tmp);
	    sizeup();
	    return;
	 }
      }	/* --- end of method List<T>::insert  ---- */

   template < class T >
      void List<T>::insert ( T val )
      {
	 if(!front()){
	    Node<T> * tmp = new Node<T>(val);
	    front(tmp);
	    endd(tmp);
	    cursor(tmp);
	    sizeup();
	    return;
	 }else{
	    Node<T> * tmp = new Node<T>(val, endd());
//	    endd()->next(tmp);   // redundant
	    endd(tmp);
	    cursor(tmp);
	    sizeup();
	    return;
	 }
      }	/* --- end of method List<T>::insert  ---- */


   template < class T >
      void List<T>::insert_end ( T val )
      {
	 insert(val);
      }

   template < class T >
      void List<T>::display ()
      {
//	 std::cout << "*********DISPLAYING ENTIRE LIST************" << std::endl;
	 if(endd() == 0){
          std::cout << "Empty List" << std::endl;
	    return;
	 }

	 cursor() = front();
	 while(cursor() != endd() ){
	    std::cout << *(cursor()->value()) << std::endl;
	    cursor(cursor()->next());
	 }
	 std::cout << *(cursor()->value()) << std::endl; //display the end


	 return ;
      }		/* -----  end of method List<T>::display  ----- */

   template < class T >
      void List<T>::find_value(T val)
      {
	 cursor() = front();
	 while(cursor() != endd()){
	    if (*(cursor()->value()) == val)
	       return;
	    else{
	       cursor(cursor()->next());
	    }
	 }
	 if(*(endd()->value()) == val)
	    return;
	 else{
	    cursor() = 0; //park the cursor when nothing is found
	 }
      }

   template<class T>
      void List<T>::remove_front(){
////	 std::cout << "Removing the Front " << front() << " End of List is "<< endd() << std::endl;

	 if(front() == 0){
//	    std::cout << "Front  is NULL " << front() << std::endl;
	    return;
	 }
	 if(front() == endd()){
//
//	    std::cout << "Front  is End " << front() << std::endl;
	    Node<T> * tmp = front();
	    delete tmp;
	    front(0);
	    endd(0);
	    cursor() = 0;
	    sizedown();
	    return;
	 }

//	 std::cout << "Front  is NOT End " << front() << std::endl;
	 Node<T> * tmp = front();
	 front(front()->next());
	 delete tmp;
	 sizedown();
	 cursor() = front();
//	 std::cout << "Next Front returned " << front() << std::endl;
	 return;
      }



   template< class T>
      void List<T>::remove_all(){

//	 std::cout << "remove_all " << __LINE__ << std::endl;
	 cursor() = front();
	 while(cursor() != endd()){
//	    std::cout << "remove_all the front line" << __LINE__ << std::endl;
	    remove_front();
	 }
//	 std::cout << "\n\nReached the End\n\n";
	 remove_front();
//	 std::cout << "\n\nDeleted Last Node\n\n";
      }

   template<class T>
      List<T>::~List<T>(){
	 remove_all();
//	 std::cout << "Deleted All*************"  << __LINE__ << std::endl;
      }

   template < class T >
      Node<T>*& List<T>::find_next_value (const T& value, Node<T>*& last_node_searched) 
      {
//	 std::cout << "IN FIND_NEXT_VALUE \n";
	 cursor(last_node_searched);
//	 std::cout << "CURSOR AT " << cursor() << " and END IS AT " << endd() << std::endl;
	 if(cursor() == endd()){
	    cursor() = 0;
	    return cursor();
	 }
//	 std::cout << "Not at the end yet!!\n";
	 cursor() = cursor()->next();
//	 std::cout << "Next Cursor " << cursor() << "\n";
//	 std::cout << "Cursor Value " << *(cursor()->value()) << "\n";
	 while(value != *(cursor()->value()) ){
	    if(cursor() == endd()){
	       if( *(endd()->value()) == value){
		  return endd();
	       }else{
		  cursor() = 0;
		  return cursor();//is this a problem.  Does setting the Node * to zero return a zero address on all platforms?
	       }
	    }
	    cursor() = cursor()->next();
//	    std::cout << "Next Cursor " << cursor() << "\n";
//	    std::cout << "CURSOR AT " << cursor() << " and END IS AT " << endd() << std::endl;
	 }
	 return cursor();
      }		/* -----  end of method List<T>::find_next_value  ----- */

   template< class T >
      T& chainlist::List<T>::operator[](unsigned int const index){
	   // std::cout << "Don't INDEX past the end of the file!  Use obj_pt->size():  Size => " << size() << "Index=> " << index << std::endl;
	 unsigned int i = 0;
	 if( index >= (unsigned int) size() ){
	    //std::cerr << "Don't INDEX past the end of the file!  Use obj_pt->size():  Size => " << size() << "Index=> " << index << std::endl;
	    //std::exit(1) ; 
	    T *dump;
	    for(i = size(); i <= index ;++i){
	       dump = new T();
	       insert(*dump);
	       delete dump;
	    };
	    return *(endd()->value());
	 }

	 if(index == 0){
	   // std::cout << "Don't INDEX past the end of the file!  Use obj_pt->size():  Size => " << size() << "Index=> " << index << std::endl;
	    cursor() = front();
	    return *(cursor()->value());
	 }


	 cursor() = front();
	 for(; i < index; ++i){
	    cursor(cursor()->next());
	 }
	 return *(cursor()->value());
      }

   template <class T>
      void List<T>::sort ( List <T> & v ) 
      {
	 const size_t n = v.size();

	 for (int gap = n/2; 0<gap; gap/=2){
	    //cout << " gap==> " << gap << " n ==> " << n << endl;  
	    for (int i = gap;i < int(n);i++){
	       //cout << "i ==> " << i << " gap==> " << gap << " n ==> " << n << endl; 
	       for (int j=i-gap;0<=j ;j-=gap ){  
		  //cout << "COMPARE::i ==> " << i << " j==> " << j << " gap==> " << gap << " n ==> " << n << endl;       
		  //cout << "***Comparing " << j+gap << " and " << j << endl; 
		  if(v[j+gap]<v[j]){
		     T temp = v[j];
		     v[j] = v[j+gap];
		     v[j+gap] = temp;
		  }   
	       }   
	    }   
	 }
      }






}		/* -----  end of namespace chainlist  ----- */
#endif /* LINKLIST_H */
